from skretrieval.retrieval import RetrievalTarget
from skretrieval.retrieval.tikhonov import two_dim_vertical_second_deriv
import xarray as xr
import numpy as np
import sasktran as sk
from skretrieval.core.radianceformat import RadianceBase, RadianceGridded
from typing import List
import pandas as pd
from copy import copy
from osiris_skretrieval.osiris_target import OsirisTarget



def polyfit_matrix(x, deg=2, ozone_cross=None, no2_cross=None, solar_spectrum=None, g12=None, g13=None):
    num_pred = deg + 1
    if ozone_cross is not None:
        num_pred += 1
    if no2_cross is not None:
        num_pred += 1
    if solar_spectrum is not None:
        num_pred += 1
    if g12 is not None:
        num_pred += 1
    if g13 is not None:
        num_pred += 1
    mat = np.zeros((len(x), num_pred))
    c = -1
    if no2_cross is not None:
        mat[:, c] = no2_cross
        c -= 1
    if ozone_cross is not None:
        mat[:, c] = ozone_cross
        c -= 1
    if solar_spectrum is not None:
        mat[:, c] = solar_spectrum
        c -= 1
    if g12 is not None:
        mat[:, c] = g12
        c -= 1
    if g13 is not None:
        mat[:, c] = g13
        c -= 1

    for i in range(deg+1):
        mat[:, i] = x**i

    return mat, np.linalg.pinv(mat)


class OSIRISNO2Retrieval(OsirisTarget):
    def __init__(self, no2_species: sk.Species, measurement_l1: RadianceGridded, state_vector_space='log', tikh_factor=0, high_alt_normalize=False,
                 ret_m_spacing=1000, ozone_cross_fwhm=None, fit_mode='doas', apriori_clim=None,
                 apriori_influence=1, no2_opticalproperty=None, o3_opticalproperty=None, include_polarization_fit=True,
                 lower_altitude_bound=8500, upper_altitude_bound=48500, retrieve_on_osiris_grid=False,
                 v6_normalization_altitude=False,
                 v6_wavelength_range=False,
                 high_alt_normalization_range=5000,
                 good_pixel_fraction=1,
                 spectral_fit_range=(434.7, 472.5),
                 force_include_solar_fit=False):
        self._spectral_fit_range = spectral_fit_range
        self._force_include_solar_fit = force_include_solar_fit
        self._no2_species = no2_species
        self._v6_normalization_altitude = v6_normalization_altitude
        self._v6_wavelength_range = v6_wavelength_range
        self._good_pixel_fraction = good_pixel_fraction
        self._fit_mode = fit_mode
        self._ret_spacing = ret_m_spacing

        self._high_alt_norm_range = high_alt_normalization_range
        self._high_alt_normalize = high_alt_normalize

        self._include_polarization = include_polarization_fit

        self._ozone_cross_fwhm = ozone_cross_fwhm

        self._no2_opticalproperty = no2_opticalproperty
        self._o3_opticalproperty = o3_opticalproperty
        super().__init__(species=no2_species, measurement_l1=measurement_l1, state_vector_space=state_vector_space,
                         tikh_factor=tikh_factor, lower_altitude_bound_m=lower_altitude_bound,
                         upper_altitude_bound_m=upper_altitude_bound,
                         ret_spacing_m=ret_m_spacing, apriori_clim=apriori_clim, apriori_influence=apriori_influence)

        if retrieve_on_osiris_grid:
            os_tan_alts = measurement_l1.tangent_locations().altitude.values
            self._retrieval_altitudes = os_tan_alts[os_tan_alts < upper_altitude_bound]
        else:
            self._retrieval_altitudes = np.arange(lower_altitude_bound, upper_altitude_bound, ret_m_spacing)

        self._set_lower_bound(measurement_l1)

    def _set_lower_bound(self, l1_data: RadianceGridded):
        good = np.zeros(len(self._retrieval_altitudes))

        # Only retrieve on altitudes above the lowest altitude that is used in y
        y_meas = self.measurement_vector(l1_data)
        good_y = ~np.isnan(y_meas['y'])

        # Add a bit for rounding differences
        good[y_meas['y_altitudes'][good_y][0] < self._retrieval_altitudes + 1e-5] = 1

        self._retrieval_altitudes = self._retrieval_altitudes[good.astype(bool)]

    def rtm_options(self):
        options = dict()
        options['calcwf'] = 2
        options['wfheights'] = self._retrieval_altitudes
        options['wfwidths'] = np.gradient(self._retrieval_altitudes)
        options['numordersofscatter'] = 50

        return options, 'no2'

    def measurement_vector_doas(self, l1_data: RadianceGridded):
        if not isinstance(l1_data, RadianceGridded):
            raise ValueError('Class OSIRISNO2Retrieval only supports data in the form RadianceGridded')

        if self._v6_wavelength_range:
            good_wavel = [434.7, 476.7]
        else:
            good_wavel = self._spectral_fit_range

        result = dict()
        result['y'] = []
        if 'wf' in l1_data.data:
            result['jacobian'] = []

        if 'radiance_noise' in l1_data.data:
            result['y_error'] = []

        # If l1_data contains FWHM, use that instead of user specified FWHM
        if 'fwhm' in l1_data.data:
            fwhm = float(l1_data.data.fwhm.mean())
        else:
            fwhm = self._ozone_cross_fwhm

        tangent_locations = l1_data.tangent_locations()

        triplet_values = l1_data.data.where((l1_data.data.wavelength < good_wavel[1]) & (l1_data.data.wavelength > good_wavel[0]), drop=True)

        good_alt_index = (tangent_locations.altitude.values >= self._retrieval_altitudes[0] - self._ret_spacing / 2) &\
                         (tangent_locations.altitude.values <= self._retrieval_altitudes[-1] + 100)

        if self._v6_normalization_altitude:
            norm_alt_index = (tangent_locations.altitude.values >= 50000) & \
                             (tangent_locations.altitude.values <= 70000)
        else:
            norm_alt_index = (tangent_locations.altitude.values >= self._retrieval_altitudes[-1] + 100) &\
                             (tangent_locations.altitude.values <= self._retrieval_altitudes[-1] + self._high_alt_norm_range)

        triplet_values['raw_radiance'] = triplet_values['radiance'].copy()
        if self._high_alt_normalize:
            triplet_values['radiance'] = triplet_values['radiance'] / (triplet_values['radiance'] / triplet_values['radiance'].mean(dim='wavelength')).isel(los=norm_alt_index).mean(dim='los')
            triplet_values = triplet_values.isel(los=good_alt_index)
        else:
            triplet_values = triplet_values.isel(los=good_alt_index)

        log_radiance = np.log(triplet_values['radiance'])

        convolved_o3 = self._o3_opticalproperty
        convolved_no2 = self._no2_opticalproperty

        no2_coeff_fit = np.zeros(len(tangent_locations.los[good_alt_index]))

        if 'wf' in triplet_values:
            log_wf = triplet_values['wf'] / triplet_values['raw_radiance']
            no2_coeff_fit_wf = np.zeros((len(tangent_locations.los[good_alt_index]), len(self._retrieval_altitudes)))

        if 'y_error' in result:
            no2_coeff_error = np.zeros(len(tangent_locations.los[good_alt_index]))
            snr = triplet_values['raw_radiance'] / triplet_values['radiance_noise']

        if not self._high_alt_normalize or self._force_include_solar_fit:
            solar_spectrum = sk.SolarSpectrum()
            solar_spectrum = solar_spectrum.irradiance(log_radiance.wavelength.values, fwhm=fwhm)
            solar_spectrum -= np.nanmean(solar_spectrum)
            solar_spectrum /= np.nanstd(solar_spectrum)

            solar_fit = solar_spectrum - np.polyval(
                np.polyfit(log_radiance.wavelength.values, solar_spectrum, deg=3),
                log_radiance.wavelength.values)
        else:
            solar_fit = None

        all_coeff_dot = []
        all_residuals = []
        all_residual_alts = tangent_locations.altitude[good_alt_index].values
        all_good_pixel_fractions = []
        for idx, los in enumerate(tangent_locations.los[good_alt_index]):
            sliced = tangent_locations.sel(los=los)

            mjd = float(l1_data.data.sel(los=los).mjd)

            good_wavel = ~np.isnan(log_radiance.isel(los=idx).values)

            ozone_cross = convolved_o3.calculate_cross_sections(sk.MSIS90(), float(sliced.latitude), float(sliced.longitude),
                                                               float(sliced.altitude), mjd, log_radiance.wavelength.values[good_wavel]).total
            ozone_cross -= np.nanmean(ozone_cross)
            ozone_cross /= np.nanstd(ozone_cross)

            no2_cross = convolved_no2.calculate_cross_sections(sk.MSIS90(), float(sliced.latitude), float(sliced.longitude),
                                                               float(sliced.altitude), mjd, log_radiance.wavelength.values[good_wavel]).total
            no2_cross -= np.nanmean(no2_cross)
            no2_cross /= np.nanstd(no2_cross)

            g12 = None
            g13 = None
            if self._include_polarization:
                data = pd.read_csv(
                    r'\\utls\OsirisDataProducts\OsirisConfigurationDatabase\cdb\ospolarization/cal_eff_mar8.dat',
                    skiprows=2, header=None, sep='\s+')

                pol_w = data.values[:, 1]
                pol_g12 = data.values[:, 3]
                pol_g13 = data.values[:, 4]

                g12 = np.interp(log_radiance.wavelength.values[good_wavel], pol_w, pol_g12)
                g13 = np.interp(log_radiance.wavelength.values[good_wavel], pol_w, pol_g13)

            X, X_pinv = polyfit_matrix(log_radiance.wavelength.values[good_wavel], deg=4, ozone_cross=ozone_cross, no2_cross=no2_cross, solar_spectrum=solar_fit, g12=g12, g13=g13)
            p_fit_coeff = X_pinv @ log_radiance.isel(los=idx).values[good_wavel]

            all_coeff_dot.append(X_pinv[-1, :])

            no2_coeff_fit[idx] = p_fit_coeff[-1]

            if 'wf' in triplet_values:
                no2_coeff_fit_wf[idx, :] = (X_pinv @ log_wf.values[good_wavel, idx, :])[-1, :]

            residual = log_radiance.isel(los=idx).values[good_wavel] - p_fit_coeff @ X.T
            if 'y_error' in result:
                # Use the fit residuals to calculate the error
                no2_coeff_error[idx] = (X_pinv @ np.diag((1/snr**2).isel(los=idx).values[good_wavel]) @ X_pinv.T)[-1, -1] * np.nanstd(residual)**2 / (np.nanmean(1/snr.isel(los=idx).values[good_wavel]**2))

            if np.nansum(good_wavel) / len(log_radiance.wavelength.values) >= self._good_pixel_fraction:
                result['y'].append(no2_coeff_fit[idx])
            else:
                result['y'].append(np.nan)
            all_good_pixel_fractions.append(np.nansum(good_wavel) / len(log_radiance.wavelength.values))

            full_shape_residual = np.ones(len(log_radiance.wavelength.values))
            full_shape_residual[good_wavel] = residual
            all_residuals.append(full_shape_residual)

            if 'wf' in triplet_values:
                result['jacobian'].append(no2_coeff_fit_wf[idx, :])

                if self._state_vector_space.lower() == 'log':
                    x = np.exp(self.state_vector())
                    result['jacobian'][-1] *= x
            if 'y_error' in result:
                result['y_error'].append(no2_coeff_error[idx])
        result['y'] = np.hstack(result['y'])
        if 'y_error' in result:
            result['y_error'] = np.hstack(result['y_error'])
        if 'jacobian' in result:
            result['jacobian'] = np.vstack(result['jacobian'])

        result['fit_residual_alts'] = all_residual_alts
        result['fit_residuals'] = np.vstack(all_residuals)
        result['y_altitudes'] = tangent_locations.altitude.values[good_alt_index]
        result['fit_residual_wavelengths'] = triplet_values.wavelength.values
        result['good_pixel_fraction'] = all_good_pixel_fractions

        return result

    def measurement_vector(self, l1_data: RadianceGridded):
        if self._fit_mode == 'spectrum':
            return self.measurement_vector_spectrum(l1_data)
        elif self._fit_mode == 'doas':
            return self.measurement_vector_doas(l1_data)
        else:
            raise ValueError()

    def measurement_vector_spectrum(self, l1_data: RadianceGridded):
        if not isinstance(l1_data, RadianceGridded):
            raise ValueError('Class OSIRISNO2Retrieval only supports data in the form RadianceGridded')

        good_wavel = [435, 475]

        result = dict()
        result['y'] = []
        if 'wf' in l1_data.data:
            result['jacobian'] = []

        if 'radiance_noise' in l1_data.data:
            result['y_error'] = []

        tangent_locations = l1_data.tangent_locations()

        triplet_values = l1_data.data.where((l1_data.data.wavelength < good_wavel[1]) & (l1_data.data.wavelength > good_wavel[0]), drop=True)

        good_alt_index = (tangent_locations.altitude.values >= self._retrieval_altitudes[0] - 100) &\
                         (tangent_locations.altitude.values <= self._retrieval_altitudes[-1] + 100)

        norm_alt_index = (tangent_locations.altitude.values >= self._retrieval_altitudes[-1] + 100) &\
                         (tangent_locations.altitude.values <= self._retrieval_altitudes[-1] + self._high_alt_norm_range)

        triplet_values['raw_radiance'] = triplet_values['radiance'].copy()
        if self._high_alt_normalize:
            triplet_values['radiance'] = triplet_values['radiance'] / triplet_values['radiance'].isel(los=norm_alt_index).mean(dim='los')
            triplet_values = triplet_values.isel(los=good_alt_index)
        else:
            triplet_values = triplet_values.isel(los=good_alt_index)

        log_radiance = np.log(triplet_values['radiance'])

        if self._ozone_cross_fwhm is not None:
            ozone_cross = sk.OpticalPropertyConvolved(sk.O3DBM(), l1_data.data.wavelength.values,
                                                      self._ozone_cross_fwhm / (2 * np.sqrt(2*np.log(2.0))))

            ozone_cross = ozone_cross.calculate_cross_sections(sk.MSIS90(), 0, 0, 25000, 54372, l1_data.data.wavelength.values).total

            ozone_cross -= np.nanmean(ozone_cross)
            ozone_cross /= np.nanstd(ozone_cross)
            ozone_cross = ozone_cross - np.polyval(np.polyfit(l1_data.data.wavelength.values, ozone_cross, deg=3),
                                                   l1_data.data.wavelength.values)

        else:
            ozone_cross = None

        X, X_pinv = polyfit_matrix(log_radiance.wavelength.values, deg=3, ozone_cross=ozone_cross)

        p_fit_coeff = X_pinv @ log_radiance.values

        log_radiance.values -= X @ p_fit_coeff

        M = X @ X_pinv

        result['y'].append(log_radiance.values.flatten())

        if 'radiance_noise' in triplet_values:
            snr = triplet_values['raw_radiance'] / triplet_values['radiance_noise']

            y_error = 1 / snr.values ** 2

            y_error += (M @ y_error)

            result['y_error'].append(y_error.flatten())

        if 'wf' in triplet_values:
            log_wf = triplet_values['wf'] / triplet_values['raw_radiance']

            log_wf.values -= np.einsum('ij,j...->i...', M, log_wf.values)

            result['jacobian'].append(log_wf.values.reshape((-1, len(self._retrieval_altitudes))))

            if self._state_vector_space.lower() == 'log':
                x = np.exp(self.state_vector())
                result['jacobian'][-1] *= x[np.newaxis, :]

        result['unstacked_y'] = np.copy(result['y'])
        result['unstacked_alts'] = tangent_locations.altitude.values[good_alt_index]
        result['unstacked_wavelength'] = triplet_values.wavelength.values
        result['y'] = np.hstack(result['y'])
        if 'y_error' in result:
            result['y_error'] = np.hstack(result['y_error'])
        if 'jacobian' in result:
            result['jacobian'] = np.vstack(result['jacobian'])

        return result
